import pandas as pd
import os
import warnings
import multiprocessing

class ParallelDataFrame():
    '''
    Run pandas apply function with multiprocessing, suitable for CPU 
    intensive operation; the execution time advantage is showing with
    large dataframe/series.
    '''

    def __init__(self, df, num_ps=None):
        '''
        param:
            df: DataFrame
            num_ps: int; optional
        '''

        cpu_counts = multiprocessing.cpu_count()
        cores = cpu_counts - 1 if cpu_counts > 1 else cpu_counts

        if num_ps:
            self.pool = multiprocessing.Pool(processes=cores)
        else:
            self.pool = multiprocessing.Pool(processes=num_ps)

        self._df = df
        
        if os.name == 'nt':
            warnings.warn("requiring multiprocessing-safe main module protection: " +
                          "if __name__ == '__main__':")        


    def apply(self, func, *args, chunksize=10000):
        '''
        param:
            func: function
            chunksize: int; default to 10000

        return: DataFrame or Series
        '''

        async_results = [self.pool.apply_async(func,
                         (self._df.iloc[i: i + chunksize], *args))
                         for i in range(0, len(self._df), chunksize)]

        df_results = [_r.get() for _r in async_results]

        return pd.concat(df_results)


    def group_apply(self, func, groupkey, *args):
        '''
        param:
            groupkey: str; list of str
            func: function

        return: DataFrame or Series
        '''

        grp_obj = self._df.groupby(groupkey)

        async_results = [self.pool.apply_async(func, (grp_obj.get_group(grp),
                         groupkey, grp, *args))
                         for grp in grp_obj.groups]

        df_results = [_r.get() for _r in async_results]

        return pd.concat(df_results)


    def close(self):

        self.pool.close()


    def __enter__(self):
        return self


    def __exit__(self, exc_type, exc, tb):
        self.close()


import os
import asyncio
import jpype
from pydtc.connection import DBClient, APIClient
from pydtc.parallelize import ParallelDataFrame

def connect(db, host, user, password, props={}, classname=None, lib_path=None, runtime_path=None):
    '''
    Interface to connect to the database

    Param:
      db: the common name of the database, e.g db2, teradata
      host: the host of the database instance
      user: user name
      passowrd: user password
      props: the java property that to be set system-wide
      classname: the classname, e.g. "com.mysql.jdbc.Driver"
      lib_path: the jdbc driver location
      runtime_path: the java runtime path if custom desired
    '''
    con = DBClient(db, host, user, password, java_props=props, classname=classname, lib_path=lib_path, runtime_path=runtime_path)
    con.connect()

    return con

def read_sql(con, sql):

    return con.read_sql(sql)

def clob_to_str(clob):
    try:
        sb = jpype.java.lang.StringBuilder()
        chars = jpype.JArray(jpype.JChar)(clob.length())
        cs = clob.getCharacterStream()
        cs.read(chars)
        sb.append(chars)
        cs.close()
        return sb.toString()
    except Exception as err:
        return str(err)[:200]

def blob_to_file(blob, file_name, save_to):
    '''
    helper func to save the blob field into file.
    Param:
      blob: the blob field
      file_name: the file name
      save_to: the folder to save the file
    '''
    try:
        out = jpype.java.io.FileOutputStream(os.path.join(save_to, file_name))
        buff = blob.getBytes(1, blob.length())
        out.write(buff)
        out.close()
        return 'Success'
    except Exception as err:
        return str(err)[:200]

def create_temp(con, sql):

    con.create_temp(sql)

def load_temp(con, sql, df, chunksize=10000):

    con.load_temp(sql, df, chunksize=chunksize)

def load_batch(con, sql, df, chunksize=10000):

    con.load_batch(sql, df, chunksize=chunksize)


# speed up pandas cpu operation with multiprocessing especially for large set.
def p_apply(func, df, chunksize=10000, cores=None):
    try:
        pdf = ParallelDataFrame(df, num_ps=cores)

        return pdf.apply(func, chunksize=chunksize)
    except:
        raise
    finally:
        pdf.close()


def p_groupby_apply(func, df, groupkey, cores=None):
    try:
        pdf = ParallelDataFrame(df, num_ps=cores)

        return pdf.group_apply(func, groupkey)
    except:
        raise
    finally:
        pdf.close()


def api_get(urls, auth=None, loop=None):
    try:
        loop = loop or asyncio.get_running_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

    try:
        api = APIClient(auth=auth, loop=loop)

        if isinstance(urls, list):
            _results = loop.run_until_complete(api.fetch_all(urls))

            results = [{url : r} for url, r in zip(urls, _results)]

        else:
            _results = loop.run_until_complete(api.fetch(urls))

            results = {urls : _results}

        return results
    except:
        raise
    finally:
        asyncio.run(api.close())


def api_update(url, data=None, method='put', auth=None, loop=None):
    try:
        loop = loop or asyncio.get_running_loop()
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)

    try:
        api = APIClient(auth=auth, loop=loop)

        results = loop.run_until_complete(api.update(url, data=data, method=method))

        return results
    except:
        raise
    finally:
        asyncio.run(api.close())


from lxml import html
from urllib.parse import urlparse
import requests


class HttpFormAuth(requests.auth.AuthBase):
    '''
    Form based Authentication Handler for Requests.
    '''

    def __init__(self, username, password):
        '''
        param:
            user: str
            password: str
        '''
        self.username = username
        self.password = password
        self.redirect_cnt = 0

    def __call__(self, r):
        '''
        The hook handler.
        '''
        r.register_hook('response', self._form_login)

        return r


    def _form_login(self, r, **kwargs):
        '''
        param:
            r: Response

        return:
            Response
        '''
        if not 300 <= r.status_code < 400 and self.redirect_cnt < 1:
            try:
                payload, url, method = self._prep_form_fields(r)
                prep = r.request.copy()
                prep.prepare(method = method, url = url, data = payload, cookies = r.cookies)
                _r = r.connection.send(prep, **kwargs)
            except:
                pass
            else:
                self.redirect_cnt += 1
                return _r
        return r


    def _prep_form_fields(self, r):
        '''
        Get the authentication form fields for sign in.
        '''
        parsed_url = urlparse(r.url)
        base_url = '{uri.scheme}://{uri.netloc}{uri.path}'.format(uri=parsed_url)

        doc = html.document_fromstring(r.text, base_url=base_url)
        form = doc.xpath('//form')[0]
        user, password = self._pick_fields(form)
        form.fields[user] = self.username
        form.fields[password] = self.password

        payload = dict(form.form_values())
        return payload, form.action or form.base_url, form.method


    def _pick_fields(self, form):
        '''
        idea from diango form authentication.
        '''
        userfield = passfield = emailfield = None
        for x in form.inputs:
            if not isinstance(x, html.InputElement):
                continue

            _type = x.type
            if _type == 'password' and passfield is None:
                passfield = x.name
            elif _type == 'email' and emailfield is None:
                emailfield = x.name
            elif _type == 'text' and userfield is None:
                userfield = x.name

        return emailfield or userfield, passfield



import os
import logging
import re
import jaydebeapi
import jpype
import pandas as pd

import asyncio
import aiohttp
from pydtc.utils import exec_time, async_retry

## dict of database software name to jdbc driver class name
driver_class = {
               'db2': 'com.ibm.db2.jcc.DB2Driver',
               'teradata': 'com.teradata.jdbc.TeraDriver',
               'sqlserver': 'com.microsoft.sqlserver.jdbc.SQLServerDriver',
               'oracle': 'oracle.jdbc.driver.OracleDriver',
               'mysql': 'com.mysql.cj.jdbc.Driver',
               'hive2': 'org.apache.hive.jdbc.HiveDriver'
               }

class DBClient():
    '''
    Class wrapping the connection to database via jdbc with batch/fast
    load capability.

    The jdbc driver jar file(s) to be supplied by the user which can be easily
    accquired and to be placed into folder jdbc_driver under the user's home
    directory.
    '''

    def __init__(self, db, host, user, password, java_props={}, classname=None, lib_path=None, runtime_path=None):
        '''
        Instance of DBCon class.

        param:
            db: str; db2|teradata|mssql|mysql|hive2 with kerberos auth etc.
            host: str; url of db server.
            user: str
            password: str
            java_props: dict; the java property that to be set system-wide, optional
            classname: str, the forclass name for the jdbc driver
            lib_path: str, default to jdbc_driver under user's home directory
            runtime_path: str; location of the jvm lib, optional       
        '''

        self.logger = logging.getLogger(__name__)

        if runtime_path:
            jvm = runtime_path
        else:
            jvm = jpype.getDefaultJVMPath()

        self._db = db
        self._host = host
        self._user = user
        self._pass = password

        self._conn = None
        self._cur = None
        self._col_prop = {}

        try:
            self._driver = driver_class[db]
        except KeyError:
            if classname:
                self._driver = classname
            else:
                raise Exception('unknown driver class name. specify like: ' +
                                'classname = "com.mysql.jdbc.Driver"')

        if lib_path:
            _lib_path = lib_path
        else:
            _lib_path = os.path.join(os.path.expanduser('~'), 'jdbc_driver')
            if not os.path.exists(_lib_path):
                os.makedirs(_lib_path)

        classes = [c for c in os.listdir(_lib_path) if c.endswith('.jar')]

        if len(classes) == 0:
            raise Exception('no jar file(s) provided in folder {}.'.format(_lib_path))

        if os.name == 'nt':
            _path = ';'.join([os.path.join(_lib_path, c) for c in classes])
        else:
            _path = ':'.join([os.path.join(_lib_path, c) for c in classes])

        args = '-Djava.class.path={}'.format(_path)
        if jpype.isJVMStarted():
            pass
        else:
            jpype.startJVM(jvm, args)
            if java_props:
                system = jpype.JClass("java.lang.System")
                for k, v in java_props.items():
                    system.setProperty(str(k), str(v))


    def connect(self):
        if self._db == 'oracle':
            connectionstring = 'jdbc:{db}:thin:@{host}'.format(db=self._db, host=self._host)
        else:
            connectionstring = 'jdbc:{db}://{host}'.format(db=self._db, host=self._host)

        try:
            if self._db == 'hive2':
                import org.apache.hadoop.security.UserGroupInformation
                import org.apache.hadoop.conf.Configuration
                conf = org.apache.hadoop.conf.Configuration()
                conf.set("hadoop.security.authentication", "Kerberos")
                org.apache.hadoop.security.UserGroupInformation.setConfiguration (conf)
                org.apache.hadoop.security.UserGroupInformation.loginUserFromKeytab(self._user, self._pass)
                self._conn = jaydebeapi.connect (self. driver, connectionstring, ['', ''], None,)
                self._conn.jconn.setAutoCommit(True)
            else:
                self._conn = jaydebeapi.connect(self._driver, connectionstring,
                                                [self._user, self._pass],
                                                None,)

                self._conn.jconn.setAutoCommit(False)
            self._cur = self._conn.cursor()

            self.logger.warning('Connected: %s', self._db.title())

        except jpype.JavaException as err:
            self.logger.error(err)
            raise

    @exec_time()
    def update_sql(self, sqlstr, errmsg = 'Update Failed'):
        '''
        param:
            sqlstr: str; sql statement, e.g. create temporary table temp (id int); delete from tbl
        '''

        try:
            stmt = self._conn.jconn.createStatement()
            if sqlstr.lower().startswith('delete'):
                stmt.execute(sqlstr)
            else:
                stmt.executeUpdate(sqlstr)
            self._conn.commit()

            stmt.close()
        except Exception:
            self.logger.exception(errmsg)
            raise

    @exec_time()
    def create_temp(self, sqlstr):
        self.update_sql(sqlstr, errmsg='Temporary table creation failed.')

    @exec_time()
    def load_batch(self, sqlstr, indata, chunksize=10000, errmsg='Batch load failed.'):
        '''
        param:
            sqlstr: str; sql statement
            indata: DataFrame; data to be inserted into the table, support int and string data type
            chunksize: int; default to 10000
        sample:
            sqlstr: insert into Demo (col1, col2) values (?,?)
        '''

        if isinstance(indata, pd.DataFrame):
            try:
                pstmt = self._conn.jconn.prepareStatement(sqlstr)

                _schema = [str(indata[c].dtype) for c in indata.columns]
                for i in range(0, len(indata), chunksize):
                    _data = indata.iloc[i: i+chunksize]
                    for j in zip(*_data.T.values.tolist()):
                        for k in range(len(j)):
                            if _schema[k].find('int') == 0:
                                pstmt.setInt(k+1, j[k])
                            elif _schema[k].find('obj') == 0:
                                pstmt.setString(k+1, j[k])

                        pstmt.addBatch()

                    pstmt.executeBatch()

                if self._db == 'hive2':
                    pass
                else:
                    self._conn.commit()

                pstmt.close()
            except Exception:
                self.logger.exception(errmsg)
                raise
        else:
            raise Exception('Input takes dataframe only.')


    @exec_time()
    def load_temp(self, sqlstr, indata, chunksize=10000):
        self.load_batch(sqlstr, indata, chunksize=10000, errmsg='Temporary table insertion failed.')


    @exec_time()
    def read_sql(self, sqlstr, custom_converters=None):
        '''
        param:
            sqlstr: str; sql statement
            custom_converters: default to use builtin.
        '''
        if custom_converters == None:
            converters = jaydebeapi._converters
        elif isinstance(custom_converters, dict):
            converters = {**jaydebeapi._converters, **custom_converters}
        else:
            raise Exception('Dictionary of column type and custom function to be provide.')


        stmt = self._conn.jconn.createStatement()
        try:
            stmt.execute(sqlstr)

            result = stmt.getResultSet()
            meta = result.getMetaData()
            column_count = meta.getColumnCount()

            rows, columns, converter_func = [], [], []

            for col in range(1, column_count+1):
                _coltyp = meta.getColumnType(col)
                _colnm = meta.getColumnName(col)
                self._col_prop[_colnm] = _coltyp
                converter = converters.get(_coltyp, jaydebeapi._unknownSqlTypeConverter)
                converter_func.append(converter)
                columns.append(_colnm)
                
            while result.next():
                row = []
                for i in range(column_count):
                    row.append(converter_func[i](result, i+1))
                rows.append(row)

            if self._db == 'hive2':
                pass
            else:
                self._conn.commit()

            if rows:
                return pd.DataFrame(rows, columns=columns)
            else:
                return pd.DataFrame(columns=columns, dtype=object)
        except Exception as e:
            self.logger.error(str(e))
            raise
        finally:
            stmt.close()


    def close(self):
        try:
            self._cur.close()
            self._conn.close()
        except Exception as e:
            self.logger.warning(e)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, tb):
        self.close()


class APIClient():
    '''
    Class wrapping connection to RESTful api utilizing aiohttp package for concurrent requests.
    '''

    def __init__(self, auth=None, loop=None, **kwargs):
        '''
        param:
            auth: requests authentication; default: None
            loop: EventLoop; default: None
        '''

        self._session = aiohttp.ClientSession(auth=auth, loop=loop, **kwargs)


    @async_retry(Exception, logger = logging.getLogger('retry'))
    async def fetch(self, url):
        async with self._session.get(url) as response:
            status = response.status
            if status == 200:
                return await response.json()
            else:
                if response.text:
                    raise Exception('Status Code: {}; Message: {}'.format(status, await response.json()))
                else:
                    raise Exception('GET Failed: {}'.format(status))


    async def fetch_all(self, urls):
        results = await asyncio.gather(
            *[self.fetch(url) for url in urls],
            return_exceptions=True
        )

        return results


    @async_retry(Exception, logger=logging.getLogger('retry'))
    async def update(self, url, data=None, method='put'):
        _requests = {'post' : self._session.post,
                     'put' : self._session.put,
                     'patch' : self._session.patch,
                     'delete' : self._session.delete
        }

        if method not in _requests:
            raise Exception('unknown action.')

        async with _requests[method](url, json=data) as response:
            status = response.status
            if status == 200:
                return await response.json()
            else:
                if response.text:
                    raise Exception('Status Code: {}; Message: {}'.format(status, await response.json()))
                else:
                    raise Exception('UPDATE Failed: {}'.format(status))


    async def close(self):
        await self._session.close()
    

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self.close()


__version__ = '0.5.1'

from .pydtc import (
    connect,
    load_temp,
    create_temp,
    read_sql,
    load_batch,
    p_groupby_apply,
    p_apply,
    api_get,
    api_update,
    clob_to_str,
    blob_to_file
)

from .utils import (
    exec_time,
    retry,
    async_retry,
    DTCTimeoutException,
    timeout
)

from .formauth import HttpFormAuth


from setuptools import find_packages, setup

## create folder for user to store the jdbc driver jars
import os
jdbc_driver = os.path.join(os.path.expanduser('~'), 'jdbc_driver')
if not os.path.exists(jdbc_driver):
    os.makedirs(jdbc_driver)

project_name = 'pydtc'
## read version from __init__.py
import re
version = re.search(r'{}\s*=\s*[\'"]([^\'"]*)[\'"]'.format('__version__'), 
                   open(project_name + '/__init__.py').read()).group(1)

setup(
  name = project_name,
  packages = ['pydtc'],
  version = version, 
  description = 'tools collection for data engineer',
  long_description = open('README.md').read(),
  long_description_content_type = "text/markdown",  
  author = 'cctester',
  author_email = 'cctester2001@gmail.com',
  url = 'https://github.com/cctester/pydtc',
  keywords = ['pandas', 'multiprocessing', 'database', 'restapi', 'requests'],
  python_requires = ">=3.5",
  install_requires = [
          'pandas',
          'JayDeBeApi',
          'JPype1 == 0.6.3',
          'asyncio',
          'aiohttp'
      ],
  classifiers = [
    'Development Status :: 3 - Alpha',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: BSD License',
    'Programming Language :: Python :: 3',
  ],
)



import time
import signal
import asyncio
from functools import wraps
from multiprocessing import TimeoutError
from multiprocessing.pool import ThreadPool

def exec_time(logger=None):
    '''
    Decorator for module execution time.
    '''

    def timeit(f):
        @wraps(f)
        def timed(*args, **kwargs):
            ts = time.time()
            result = f(*args, **kwargs)
            te = time.time()
            if logger:
                logger.debug('[%r] execution time: %2.2f seconds' %
                             (f.__name__, (te-ts)))
            else:
                print('[%r] execution time: %2.2f seconds' %
                      (f.__name__, (te-ts)))
            return result
        return timed
    return timeit


def retry(*exceptions, retries=4, delay=4, logger=None):
    '''
    Decorator for function retry
    '''

    def deco_retry(f):
        @wraps(f)
        def inner(*args, **kwargs):
            mtries, mdelay = retries, delay
            while mtries > 0:
                try:
                    return f(*args, **kwargs)
                except exceptions as e:                    
                    msg = '%s; %d tries left, Retrying in %d seconds...' % (str(e), mtries, mdelay)
                    mtries -= 1

                    if logger:
                        logger.warning(msg)
                    else:
                        print(msg)

                    time.sleep(mdelay)

            return f(*args, **kwargs)
        inner.__name__ = f.__name__
        return inner
    return deco_retry


def async_retry(*exceptions, retries=4, delay=4, logger=None):
    '''
    Decorator for async function retry.
    '''

    def deco_retry(f):
        @wraps(f)
        async def inner(*args, **kwargs):
            mtries, mdelay = retries, delay

            while mtries > 0:
                try:
                    return await f(*args, **kwargs)
                except exceptions as e:
                    msg = '{}; {} tries left, Retrying in {} seconds...'.format(str(e), mtries, mdelay)
                    mtries -= 1

                    if logger:
                        logger.warning(msg)
                    else:
                        print(msg)

                    await asyncio.sleep(mdelay)

            return await f(*args, **kwargs)

        return inner
    return deco_retry

class DTCTimeoutException(Exception):
    def __init__(self, seconds):
        self.message = 'timeout after {} seconds'.format(seconds)
        super().__init__(self.message)

def timeout(s, handler=None):
    '''
    Timeout decorator.
    ref: https://anonbadger.wordpress.com/2018/12/15/python-signal-handlers-and-exceptions/
    param: 
        s:  time in seconds the function run before terminated.
        handler:  optional, for unix like system only, define custom function handling the timeout event.
    '''

    def deco_timeout(f):
        if hasattr(signal, "SIGALRM"):
            '''
            for unix like system.
            '''
            @wraps(f)
            def inner(*args, **kwargs):
                _handler = lambda _s, _f: (_ for _ in ()).throw(DTCTimeoutException(s)) if handler == None else handler

                signal.signal(signal.SIGALRM, _handler)
                signal.alarm(s)
                try:
                    return f(*args, **kwargs)
                except:
                    raise
                finally:
                    signal.alarm(0)
                
            return inner

        else:
            '''
            for others systems without SIGALRM, use multiprocessing instead.
            '''

            def wrapper(*args, **kwargs):
                pool = ThreadPool(processes=1)
                results = pool.apply_async(f, args, kwargs)
                pool.close()
                try:
                    return results.get(s)
                except TimeoutError:
                    raise DTCTimeoutException(s)
                finally:
                    pool.terminate() ## thread seems still running. todo...

            return wrapper

    return deco_timeout